'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var animationFrameChunkScheduler = typeof window !== 'undefined' &&
    typeof window.requestAnimationFrame === 'function' &&
    typeof window.cancelAnimationFrame === 'function' ?
    {
        request: function (fn) { return window.requestAnimationFrame(fn); },
        cancel: function (token) { return window.cancelAnimationFrame(token); }
    } :
    null;

var idleCallbackChunkScheduler = typeof window !== 'undefined' &&
    typeof window.requestIdleCallback === 'function' &&
    typeof window.cancelIdleCallback === 'function' ?
    {
        request: function (fn) { return window.requestIdleCallback(fn); },
        cancel: function (token) { return window.cancelIdleCallback(token); }
    } :
    null;

var immediateChunkScheduler = typeof setImmediate === 'function' &&
    typeof clearImmediate === 'function' ?
    {
        request: function (fn) { return setImmediate(fn); },
        cancel: function (token) { return clearImmediate(token); }
    } :
    null;

var timeoutChunkScheduler = {
    request: function (fn) { return setTimeout(fn, 0); },
    cancel: function (token) { return clearTimeout(token); }
};

function getChunkScheduler(type) {
    var chunkScheduler = null;
    switch (type) {
        case 'animationFrame':
            chunkScheduler = animationFrameChunkScheduler;
            break;
        case 'idleCallback':
            chunkScheduler = idleCallbackChunkScheduler;
            break;
        case 'immediate':
            chunkScheduler = immediateChunkScheduler;
            break;
        case 'timeout':
            chunkScheduler = timeoutChunkScheduler;
            break;
        case 'auto':
            chunkScheduler =
                idleCallbackChunkScheduler ||
                    animationFrameChunkScheduler ||
                    immediateChunkScheduler;
            break;
        default:
            chunkScheduler = type;
    }
    return chunkScheduler || timeoutChunkScheduler;
}

var now = typeof performance === 'object' && typeof performance.now === 'function' ?
    function () { return performance.now(); } :
    Date.now;
var microtaskPromise = Promise.resolve();
function microtask(fn) {
    microtaskPromise.then(fn);
}

function createScheduler(_a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.chunkScheduler, chunkSchedulerType = _c === void 0 ? 'auto' : _c, _d = _b.chunkBudget, chunkBudget = _d === void 0 ? 10 : _d;
    var pendingTasks = new Map();
    var chunkScheduler = getChunkScheduler(chunkSchedulerType);
    var chunkSchedulerToken = null;
    var tasksOrder = [];
    function chunk() {
        chunkSchedulerToken = null;
        var checkBudget = false;
        var startTime = now();
        var restBudget = chunkBudget;
        var nextTasksOrder = [];
        while (tasksOrder.length > 0) {
            var promise = tasksOrder.shift();
            var task = pendingTasks.get(promise);
            var taskExecuted = false;
            if (checkBudget && restBudget < task.prevUnitElapsedTime) {
                nextTasksOrder.push(promise);
            }
            else {
                checkBudget = true;
                try {
                    var _a = task.nextUnit(task.prevUnitResult), nextUnit = _a.next, result = _a.result;
                    taskExecuted = true;
                    if (nextUnit === null) {
                        pendingTasks.delete(promise);
                        task.resolve(result);
                    }
                    else {
                        task.nextUnit = nextUnit;
                        task.prevUnitResult = result;
                        tasksOrder.push(promise);
                    }
                }
                catch (e) {
                    pendingTasks.delete(promise);
                    task.reject(e);
                }
            }
            var elapsedTime = now() - startTime;
            if (taskExecuted) {
                task.prevUnitElapsedTime = elapsedTime;
            }
            restBudget -= elapsedTime;
            startTime += elapsedTime;
        }
        if (nextTasksOrder.length > 0) {
            tasksOrder = nextTasksOrder;
            chunkSchedulerToken = chunkScheduler.request(chunk);
        }
    }
    return {
        runTask: function (unit) {
            var task;
            var promise = new Promise(function (resolve, reject) {
                task = {
                    nextUnit: unit,
                    prevUnitResult: undefined,
                    prevUnitElapsedTime: 0,
                    resolve: resolve,
                    reject: reject
                };
            });
            pendingTasks.set(promise, task);
            microtask(function () {
                if (!pendingTasks.has(promise)) {
                    return;
                }
                tasksOrder.push(promise);
                if (tasksOrder.length === 1) {
                    chunkSchedulerToken = chunkScheduler.request(chunk);
                }
            });
            return promise;
        },
        abortTask: function (promise) {
            if (pendingTasks.delete(promise)) {
                var taskOrderIdx = tasksOrder.indexOf(promise);
                if (taskOrderIdx > -1) {
                    tasksOrder.splice(taskOrderIdx, 1);
                    if (tasksOrder.length === 0 && chunkScheduler.cancel && chunkSchedulerToken !== null) {
                        chunkScheduler.cancel(chunkSchedulerToken);
                        chunkSchedulerToken = null;
                    }
                }
            }
        }
    };
}

exports.createScheduler = createScheduler;
